from ConfigParser import (
    ConfigParser,
    NoSectionError)
from contextlib import contextmanager
import errno
import os
from pwd import getpwnam
import shutil
import subprocess
from tempfile import mkdtemp
import yaml

from Cheetah.Template import Template

import hookenv


CHARM_DIR = os.getcwd()
HOME = '/home/ubuntu'
PROJECT_DIR = os.path.join(HOME, 'juju-reports')
INI = 'production.ini'
REVISION_INFO_FILE = os.path.join(HOME, 'last-source-revision')


def get_ini_path(source_path, base_filename=INI):
    return os.path.join(source_path, base_filename)


def get_ini(source_path):
    config = ConfigParser()
    config.read(get_ini_path(source_path))
    return config


def get_template_path(name, source_path):
    return os.path.join(source_path, 'resources/{}.template'.format(name))


@contextmanager
def update_ini(source_path, base_filename):
    template_path = get_template_path(base_filename, source_path)
    output_path = get_ini_path(source_path, base_filename)
    if not os.path.exists(template_path):
        template_path = output_path
    config = ConfigParser()
    config.read(template_path)
    yield config
    with open(output_path, 'w') as out_file:
        config.write(out_file)


def install_ini(mongo_url, gh_token, cidata_url, token,
                error_email, error_email_from, cloud_health_failure_threshold,
                current_conditions_list, jujuci_url, jujuci_username,
                jujuci_password, source_path):
    with update_ini(source_path, INI) as config:
        config.set('app:main', 'mongo.url', mongo_url)
        config.set('app:main', 'gh_token', gh_token)
        config.set('app:main', 'jujuci.url', jujuci_url)
        config.set('app:main', 'jujuci.username', jujuci_username)
        config.set('app:main', 'jujuci.password', jujuci_password)
        config.set('app:main', 'cidata.url', cidata_url)
        config.set('app:main', 'charm_bundle_test_token', token)
        config.set('app:main', 'home', HOME)
        config.set('app:main', 'error_email', error_email)
        config.set('app:main', 'error_email_from', error_email_from)
        config.set(
            'app:main', 'cloud_health_failure_threshold',
            cloud_health_failure_threshold)
        config.set(
            'app:main', 'current_conditions_list',
            current_conditions_list)
    with update_ini(source_path, 'test.ini') as test_config:
        test_config.set('app:main', 'mongo.url', mongo_url)
    return config


def configure_s3(config, source_path):
    template_path = get_template_path('juju-qa.s3cfg', source_path)
    try:
        template_cm = open(template_path)
    except IOError as e:
        if e.errno == errno.ENOENT:
            hookenv.log('Skipping s3 config due to missing template.')
            return
    with template_cm as template_file:
        template = template_file.read()
    config_path = os.path.join(source_path, 'juju-qa.s3cfg')
    s3config = template.format(access_key=config['aws-access-key'],
                               secret_key=config['aws-secret-key'])
    update_file(config_path, s3config)
    home_stat = os.stat(HOME)
    data_path = os.path.join(HOME, 'juju-qa-data')
    for subdir in ['', 'cbtest', 'charm-test', 'industrial-test']:
        subdir_path = os.path.join(data_path, subdir)
        ensure_dir(subdir_path)
        os.chown(subdir_path, home_stat.st_uid, home_stat.st_gid)


def ensure_dir(path):
    try:
        os.mkdir(path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


def get_configured_port(ini):
    return ini_get(ini, 'server:main', 'port')


def set_port(ini, open_port=True):
    port = get_configured_port(ini)
    if port is None:
        return
    if open_port:
        hookenv.open_port(port)
    else:
        hookenv.close_port(port)


def call_script(script_path, source_path, root=False):
    env = os.environ.copy()
    ini = get_ini_path(source_path)
    env['HOME'] = HOME
    env['INI'] = ini
    if root:
        args = []
    else:
        args = ['su', 'ubuntu', '-p', '-c']

    args.append(script_path)
    subprocess.check_call(args, env=env, cwd=source_path)
    return env


def stop():
    try:
        call_script('scripts/stop', PROJECT_DIR, root=True)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise


def start(source_path):
    call_script('scripts/run', source_path)


def open_secret(path, uid=None, gid=None):
    fd = os.open(path, os.O_WRONLY | os.O_TRUNC | os.O_CREAT, 0o600)
    if uid is not None:
        os.fchown(fd, uid, gid)
    return os.fdopen(fd, 'w')


def configure_lp(ssh_key):
    if ssh_key == '':
        return False
    with open_secret('/root/.ssh/lp_rsa') as key_file:
        key_file.write(ssh_key)
    pwd = getpwnam('ubuntu')
    ubuntu_path = os.path.join(HOME, '.ssh/lp_rsa')
    with open_secret(ubuntu_path, pwd.pw_uid, pwd.pw_gid) as key_file:
        key_file.write(ssh_key)
    # Login to LP
    call_bzr(['lp-login', 'juju-qa-bot'])
    return True


def call_bzr(args):
    args = ['bzr', '--no-aliases'] + args
    return subprocess.check_call(args)


def ensure_repository(pwd):
    """Ensure that there is a repository present.

    :param pwd: The /etc/passwd entry for the user who should own the
        repository
    """
    # Grab the code
    repository = os.path.join(HOME, 'juju-reports-repo')
    try:
        os.mkdir(repository)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise
        pass
    else:
        call_bzr(['init-repo', repository])
        os.chown(repository, pwd.pw_uid, pwd.pw_gid)
    return repository


def update_source(source_url, revno, in_place):
    pwd = getpwnam('ubuntu')
    if not in_place:
        repository = ensure_repository(pwd)
        target_dir = mkdtemp(prefix='juju-reports-%s-' % revno, dir=repository)
        os.chown(target_dir, pwd.pw_uid, pwd.pw_gid)
    else:
        target_dir = PROJECT_DIR
    revno = 'revno:%s' % revno
    if not in_place or not os.path.exists(target_dir):
        call_bzr(['init', target_dir])
    else:
        try:
            real_path = os.readlink(target_dir)
        except OSError as e:
            if e.errno != errno.EINVAL:
                raise
        else:
            if in_place:
                call_bzr(['reconfigure', '--standalone', real_path])
                os.unlink(target_dir)
                os.rename(real_path, target_dir)
                # Force rebuilding the virtualenv, just this once.
                make('distclean', target_dir)

    os.chown(target_dir, pwd.pw_uid, pwd.pw_gid)

    # Make it easier to develop on a deploy
    call_bzr(['bind', '-d', target_dir, source_url])

    bzr_cmd = ['pull', '--overwrite', source_url, '-r', revno, '-d',
               target_dir, '--remember', '--local']
    call_bzr(bzr_cmd)
    return target_dir


def make(target, cwd):
    """Run a make command.

    :param target: The make target to make
    :param cwd: The directory to run make in.
    """
    subprocess.check_call(['make'] + [target], cwd=cwd)


def install_production(dist_clean, develop, source_path):
    make('deploydeps', source_path)
    if dist_clean:
        make('distclean', source_path)
    target = 'install' if develop else 'prodinstall'
    make(target, cwd=source_path)
    subprocess.check_call(['apt-get', 'autoremove', '--yes'])


def ensure_deleted(path):
    """Delete the specified file, if present."""
    try:
        os.unlink(path)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise


def update_cronjob(contents=None):
    path = '/etc/cron.d/ubuntu'
    if contents is None:
        ensure_deleted(path)
    else:
        with open(path, 'w') as outfile:
            outfile.write(contents)


def install_cronjob(interval, error_email):
    t = Template(file='templates/crontab.tmpl')
    t.interval = interval
    if error_email != '':
        t.error_email = 'MAILTO=%s' % error_email
    else:
        t.error_email = ''
    t.reports_home = PROJECT_DIR
    t.logs = os.path.join(HOME, 'logs')
    t.lockdir = os.path.join(HOME, 'locks')
    ensure_dir(t.lockdir)
    update_cronjob(str(t))


def ini_get(ini, section, key):
    try:
        return ini.get(section, key)
    except NoSectionError:
        return


def get_mongo_url():
    """Use the 'database' relation settings to build a mongodb url.

    If no mongodb units are present, return an empty string.
    """
    host_ports = []
    replset = None
    for relation_id in hookenv.relation_ids('database'):
        for unit in hookenv.related_units(relation_id):
            relation_data = hookenv.relation_get(rid=relation_id, unit=unit)
            # Consider configuration incomplete until port is supplied.
            if (relation_data is None or 'port' not in relation_data
                    or 'hostname' not in relation_data):
                continue
            if replset is None:
                replset = relation_data.get('replset')
            # All units should be members of the same replication set.
            if replset is not None and replset != relation_data['replset']:
                raise AssertionError('DB instances are from different sets!')
            host_ports.append(
                '%(hostname)s:%(port)s' % relation_data)
    if len(host_ports) == 0:
        return ''
    replica_str = ''
    if replset is not None:
        replica_str = '?replicaSet=%s' % replset
    return 'mongodb://%s/%s' % (','.join(host_ports), replica_str)


def update_website(ini):
    relation_settings = {
        'port': get_configured_port(ini),
        'hostname': hookenv.unit_private_ip(),
    }
    for relation_id in hookenv.relation_ids('website'):
        hookenv.relation_set(relation_id, relation_settings)


def update_file(path, new_contents):
    """Write file only if the contents are changed."""
    try:
        file_context = open(path, 'r')
    except IOError as e:
        if e.errno == errno.ENOENT:
            old_contents = None
        else:
            raise
    else:
        with file_context as fileobj:
            old_contents = fileobj.read()
    if old_contents == new_contents:
        return
    with open(path, 'w') as fileobj:
        fileobj.write(new_contents)


def finish_update_source(source_path, branch_url, revno):
    """Finish updating the source.

    For in-place installs, this just updates the revision info.  For other
    installs, this additionally creates a symlink from PROJECT_DIR to the
    actual install location.
    """
    if source_path != PROJECT_DIR:
        try:
            old_sources = os.readlink(PROJECT_DIR)
        except OSError as err:
            if err.errno == errno.EINVAL:
                shutil.rmtree(PROJECT_DIR)
                old_sources = None
            elif err.errno == errno.ENOENT:
                old_sources = None
            else:
                raise
        else:
            if old_sources == source_path:
                return
            os.remove(PROJECT_DIR)
        os.symlink(source_path, PROJECT_DIR)
        if old_sources is not None:
            shutil.rmtree(old_sources)
    with open(REVISION_INFO_FILE, 'w') as f:
        yaml.dump((branch_url, revno), f)


def get_last_revision():
    try:
        file_context = open(REVISION_INFO_FILE, 'r')
    except IOError as e:
        if e.errno == errno.ENOENT:
            return None, None
        else:
            raise
    with file_context:
        return yaml.load(file_context)


def update_from_config():
    ini = get_ini(PROJECT_DIR)
    mongo_url = get_mongo_url()
    config = hookenv.config()
    set_port(ini, False)
    has_lp_key = configure_lp(config['lp-key'])
    if config['source'] == '':
        hookenv.log('Incomplete config: source')
        return
    if not has_lp_key and config['source'].startswith('bzr+ssh'):
        hookenv.log('Incomplete config: ssh URL but no LP key')
        return
    needs_install_update = (
        get_last_revision() != (config['source'], config['revno']))
    if needs_install_update:
        source_path = update_source(
            config['source'], config['revno'], config['develop-install'])
    else:
        source_path = PROJECT_DIR
    oauth_path = os.path.join(source_path, 'jujureports/lp_credentials.txt')
    update_file(oauth_path, config['lp-oauth'] + '\n')
    if needs_install_update:
        install_production(
            config['dist-clean'], config['develop-install'], source_path)
    if mongo_url == '':
        hookenv.log('No mongodb set up.')
        if needs_install_update:
            finish_update_source(source_path, config['source'],
                                 config['revno'])
        return
    ini = install_ini(mongo_url, config['gh-token'],
                      config['cidata-url'], config['charm-bundle-test-token'],
                      config['error-email'], config['error-email-from'],
                      config['cloud-health-failure-threshold'],
                      config['current-conditions-list'],
                      config['jujuci-url'], config['jujuci-user'],
                      config['jujuci-pass'], source_path)
    configure_s3(config, source_path)
    unset = [key for key in [
        'lp-key', 'lp-oauth', 'gh-token', 'charm-bundle-test-token',
        'error-email', 'error-email-from'
    ] if config[key] == '']
    if len(unset) > 0:
        hookenv.log('Not setting up cron because {} not set.'.format(
            ', '.join(unset)))
        update_cronjob(None)
    elif config['cron-interval'] == -1:
        hookenv.log('Cron disabled by cron-interval = -1')
        update_cronjob(None)
    else:
        install_cronjob(config['cron-interval'], config['error-email'])
    stop()
    if needs_install_update:
        finish_update_source(source_path, config['source'], config['revno'])
    start(source_path)
    set_port(ini, True)
    update_website(ini)
